@@ -7,6 +7,8 @@ module Option = BatOption
 exception Error of string * exn
 exception Sqlite_error of string * Sqlite3.Rc.t
 
+let tx_id_counter = ref 0
+
 let curr_thread_id () = Thread.id (Thread.self ())
 
 let raise_thread_error ?msg expected =
@@ -183,27 +185,28 @@ struct
 
   let failwithfmt fmt = ksprintf failwith fmt
 
-  let error s =
-    failwithfmt "Sqlexpr_sqlite error: bad data (expected %s)" s
+  let error s x =
+    failwithfmt "Sqlexpr_sqlite error: bad data (expected %s but got %s)"
+      s (Sqlite3.Data.to_string_debug x)
 
   let text = function
       TEXT s | BLOB s -> s
     | INT n -> Int64.to_string n
     | FLOAT f -> string_of_float f
-    | _ -> error "text"
+    | x -> error "text" x
 
-  let blob = function BLOB s | TEXT s -> s | _ -> error "blob"
+  let blob = function BLOB s | TEXT s -> s | x -> error "blob" x
 
-  let int = function INT n -> Int64.to_int n | _ -> error "int"
-  let int32 = function INT n -> Int64.to_int32 n | _ -> error "int"
-  let int64 = function INT n -> n | _ -> error "int"
+  let int = function INT n -> Int64.to_int n | x -> error "int" x
+  let int32 = function INT n -> Int64.to_int32 n | x -> error "int" x
+  let int64 = function INT n -> n | x -> error "int" x
 
-  let bool = function INT 0L -> false | INT _ -> true | _ -> error "int"
+  let bool = function INT 0L -> false | INT _ -> true | x -> error "int" x
 
   let float = function
       INT n -> Int64.to_float n
     | FLOAT n -> n
-    | _ -> error "float"
+    | x -> error "float" x
 
   let maybe f = function
       NULL -> None
@@ -226,6 +229,12 @@ let profile_ch =
             (Unix.getenv "OCAML_SQLEXPR_PROFILE"))
   with Not_found -> None
 
+let raw_profile_ch =
+  try
+    Some (open_out_gen [Open_append; Open_creat; Open_binary] 0o644
+            (Unix.getenv "OCAML_SQLEXPR_LOG"))
+  with Not_found -> None
+
 let profile_uuid =
   let uuid =
     sprintf "%s %d %d %g %s %g"
@@ -296,6 +305,16 @@ struct
             [ "name"; Digest.to_hex (Digest.string sql); "portal"; " " ]
           in profile_op "execute" details f
 
+  let profile_execute_sql sql ?(full_sql=sql) ?params f =
+    let param_str = match params with
+        None -> ""
+      | Some l -> String.concat "\t" (List.rev_map string_of_param l)
+    in
+      Option.may
+        (fun ch -> fprintf ch "%s\t%s\n%!" (String.escaped full_sql) param_str)
+        raw_profile_ch;
+      profile_execute_sql sql ?params f
+
   let profile_prepare_stmt sql f =
     match profile_ch with
         None -> f ()
@@ -308,6 +327,9 @@ end
 module type POOL =
 sig
   type 'a result
+
+  module TLS : Sqlexpr_concurrency.THREAD_LOCAL_STATE with type 'a t := 'a result
+
   type db
   type stmt
   val open_db : ?init:(Sqlite3.db -> unit) -> string -> db
@@ -323,9 +345,12 @@ sig
   val raise_error :
     stmt -> ?sql:string -> ?params:Sqlite3.Data.t list -> ?errmsg:string ->
     Sqlite3.Rc.t -> 'a result
-  val unsafe_execute : db -> string -> unit result
+
+  val unsafe_execute : db -> ?retry_on_busy:bool -> string -> unit result
   val borrow_worker : db -> (db -> 'a result) -> 'a result
   val steal_worker : db -> (db -> 'a result) -> 'a result
+
+  val transaction_key : db -> unit TLS.key
 end
 
 module WT = Weak.Make(struct
@@ -343,7 +368,9 @@ type single_worker_db =
   stmt_cache : Stmt_cache.t;
 }
 
-module IdentityPool(M: THREAD) =
+let identity_pool_transaction_key_table = Hashtbl.create 13
+
+module IdentityPool(M: THREAD with type 'a key = 'a Lwt.key) =
 struct
   module Lwt = M
   open Lwt
@@ -357,6 +384,17 @@ struct
 
   let get_handle db = db.handle
 
+  let transaction_key =
+    let t = identity_pool_transaction_key_table in
+      (fun db ->
+         try
+           Hashtbl.find t db.id
+         with Not_found ->
+           let k = M.new_key () in
+             Hashtbl.add t db.id k;
+             register_finaliser (fun db -> Hashtbl.remove t db.id; return ()) db;
+             k)
+
   let handle db =
     if db.thread_id <> curr_thread_id () then
       try_lwt (raise_thread_error ~msg:"in IdentityPool.handle" db.thread_id)
@@ -387,7 +425,8 @@ struct
       with Not_found ->
         let m = M.create_recursive_mutex () in
           Hashtbl.add mutex_tbl id m;
-          Gc.finalise (fun _ -> Hashtbl.remove mutex_tbl id) db;
+          M.register_finaliser
+            (fun _ -> Hashtbl.remove mutex_tbl id; return ()) db;
           m
 
   let make handle =
@@ -414,17 +453,17 @@ struct
           sprintf "%s with params %s" msg (string_of_params (List.rev params))
     in M.fail (Error (msg, Sqlite_error (msg, errcode)))
 
-  let rec run ?stmt ?sql ?params db f x = match f x with
+  let rec run ?(retry_on_busy=false) ?stmt ?sql ?params db f x = match f x with
       Sqlite3.Rc.OK | Sqlite3.Rc.ROW | Sqlite3.Rc.DONE as r -> return r
-    | Sqlite3.Rc.BUSY | Sqlite3.Rc.LOCKED ->
-        M.sleep 0.010 >> run ?sql ?stmt ?params db f x
+    | Sqlite3.Rc.BUSY when retry_on_busy ->
+        M.sleep 0.010 >> run ~retry_on_busy ?sql ?stmt ?params db f x
     | code ->
         let errmsg = Sqlite3.errmsg db in
           Option.may (fun stmt -> ignore (Stmt.reset stmt)) stmt;
           raise_error db ?sql ?params ~errmsg code
 
-  let check_ok ?stmt ?sql ?params db f x =
-    lwt _ = run ?stmt ?sql ?params db f x in return ()
+  let check_ok ?retry_on_busy ?stmt ?sql ?params db f x =
+    lwt _ = run ?retry_on_busy ?stmt ?sql ?params db f x in return ()
 
   let prepare db f (params, nparams, sql, stmt_id) =
     lwt dbh = handle db in
@@ -466,7 +505,7 @@ struct
       iteri
         (fun n v -> check_ok ~sql ~stmt dbh (Stmt.bind stmt (nparams - n)) v)
         params >>
-      profile_execute_sql sql ~params
+      profile_execute_sql ~full_sql:sql sql ~params
         (fun () ->
            try_lwt
              f stmt sql params
@@ -490,12 +529,14 @@ struct
   let reset stmt = ignore (Stmt.reset stmt); return ()
   let row_data stmt = return (Stmt.row_data stmt)
 
-  let unsafe_execute db sql =
+  let unsafe_execute db ?retry_on_busy sql =
     lwt dbh = handle db in
-      check_ok ~sql dbh (Sqlite3.exec dbh) sql
+      check_ok ?retry_on_busy ~sql dbh (Sqlite3.exec dbh) sql
 
   let raise_error stmt ?sql ?params ?errmsg errcode =
     raise_error (Stmt.db_handle stmt) ?sql ?params ?errmsg errcode
+
+  module TLS = M
 end
 
 module type S =
@@ -533,7 +574,11 @@ sig
   val select_one_f : db -> ('a -> 'b result) -> ('c, 'a, 'b result) expression -> 'c
   val select_one_f_maybe : db -> ('a -> 'b result) ->
     ('c, 'a, 'b option result) expression -> 'c
-  val transaction : db -> (db -> 'a result) -> 'a result
+
+  val transaction :
+    db -> ?kind:[`DEFERRED | `IMMEDIATE | `EXCLUSIVE] ->
+    (db -> 'a result) -> 'a result
+
   val fold :
     db -> ('a -> 'b -> 'a result) -> 'a -> ('c, 'b, 'a result) expression -> 'c
   val iter : db -> ('a -> unit result) -> ('b, 'a, unit result) expression -> 'b
@@ -699,32 +744,61 @@ struct
 
   let new_tx_id =
     let pid = Unix.getpid () in
-    let n = ref 0 in
-      fun () -> incr n; sprintf "__sqlexpr_sqlite_tx_%d_%d" pid !n
+      fun () ->
+        (* No allocation here, so cannot have a context change until the
+         * sprintf, at least in native code. *)
+        let n = !tx_id_counter in
+          incr tx_id_counter;
+          if !tx_id_counter < 0 then tx_id_counter := 0;
+          sprintf "__sqlexpr_sqlite_tx_%d_%d" pid n
 
-  let unsafe_execute db fmt =
-    ksprintf (POOL.unsafe_execute db) fmt
+  let unsafe_execute db ?retry_on_busy fmt =
+    ksprintf (POOL.unsafe_execute db ?retry_on_busy) fmt
 
-  let unsafe_execute_prof text db fmt =
+  let unsafe_execute_prof text db ?retry_on_busy fmt =
     ksprintf
       (fun sql ->
          profile_prepare_stmt text (fun () -> return ()) >>
-         profile_execute_sql text (fun () -> POOL.unsafe_execute db sql))
+         profile_execute_sql ~full_sql:sql text (fun () -> POOL.unsafe_execute db ?retry_on_busy sql))
       fmt
 
-  let transaction db f =
+  (* wrap in BEGIN/COMMIT only for outermost txs *)
+  let outer_transaction_wrap ~kind f db =
+    match POOL.TLS.get (POOL.transaction_key db) with
+        Some _ -> f db
+      | None ->
+          let tx_kind = match kind with
+              `DEFERRED -> "DEFERRED"
+            | `IMMEDIATE -> "IMMEDIATE"
+            | `EXCLUSIVE -> "EXCLUSIVE"
+          in
+            unsafe_execute_prof ~retry_on_busy:true "BEGIN" db "BEGIN %s" tx_kind >>
+            match_lwt
+              try_lwt
+                lwt x = POOL.TLS.with_value
+                          (POOL.transaction_key db) (Some ()) (fun () -> f db)
+                in
+                  return (`OK x)
+              with exn -> return (`EXN exn)
+            with
+              | `OK x -> unsafe_execute_prof ~retry_on_busy:true
+                           "COMMIT" db "COMMIT" >> return x
+              | `EXN exn -> unsafe_execute_prof "ROLLBACK" db "ROLLBACK" >> fail exn
+
+  let transaction db ?(kind = `DEFERRED) f =
     let txid = new_tx_id () in
       POOL.steal_worker db
-        (fun db ->
-           unsafe_execute db "SAVEPOINT %s" txid >>
+        (outer_transaction_wrap ~kind begin fun db ->
+           unsafe_execute_prof "SAVEPOINT" db "SAVEPOINT %s" txid >>
            try_lwt
              lwt x = f db in
                unsafe_execute_prof "RELEASE" db "RELEASE %s" txid >>
                return x
            with e ->
              unsafe_execute_prof "ROLLBACK" db "ROLLBACK TO %s" txid >>
-             unsafe_execute db "RELEASE %s" txid >>
-             fail e)
+             unsafe_execute_prof "RELEASE" db "RELEASE %s" txid >>
+             fail e
+         end)
 
   let fold db f init expr =
     do_select
@@ -765,7 +839,7 @@ struct
       expr.statement
 end
 
-module Make(M : THREAD) = struct
+module Make(M : THREAD with type 'a key = 'a Lwt.key) = struct
   module Id = IdentityPool(M)
   include Make_gen(M)(Id)
   let make = Id.make
